/*
** MPASource. version 0.1
** (c) 2002 - Ernst Pech
** AviSynth Plugin for AviSynth 2.0x based on mpg123, version from lame 
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "internal.h"
#include "interface.h"
#include <fstream.h>

class MPASource : public IClip {
	VideoInfo vi;
	
// static variables for mpg123
	ifstream mpfile;					//mpa-file
	struct mpstr_tag mp;				//interface struct

// static variables for mpasource
	 unsigned char inbuf[16384];	//input buffer for mpa-file-data
	 char outbuf[8192];			//output buffer for pcm samples
	 int size;					//size of decoded stream in bytes
	 long last_sample;			//last requested sample
	 int last_byte;				//last used outbuf byte
	 int channels;				//number of channels (1 or 2)
	 short max_volume;			//for normalizing
	 bool normalize;

public:
    MPASource(const char filename[], const bool _normalize, IScriptEnvironment* env);
	~MPASource();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    bool __stdcall GetParity(int n);
    void __stdcall GetAudio(void* buf, int start, int count, IScriptEnvironment* env);
    const VideoInfo& __stdcall GetVideoInfo();

	int __stdcall GetSample( void * pcm, int sample_offset, long sample_number);

private:
	static __inline short Saturate(int n) {
		if (n <= -32768) return -32768;
		if (n >= 32767) return 32767;
		return (short)n;
	}
};


int __stdcall MPASource::GetSample( void * pcm, int sample_offset, long sample_number) {
// gets a sample and puts it into sample_data with offset 

	int len;	//length of input buffer
	long byte_total;	//output PCM bytes
	int ret;
	short cur_volume;

	if (pcm == NULL) 	// first call -- counting samples, calculating max_volume for normalizing
	{	
		InitMP3(&mp);
		max_volume = 0;
		normalize ? mp.quick_dirty = false : mp.quick_dirty = true;	//fast seek, not when normalizing
		byte_total = 0;
		while ( !mpfile.eof() )
		{
			mpfile.read(inbuf,16384);
			len = mpfile.gcount();
			ret = decodeMP3(&mp,inbuf,len,outbuf,8192,&size);
			if (ret==MP3_OK) {	//decode first output buffer if MP3_OK
				byte_total = byte_total+size;
				for (int j = 0; j<size/2; j++) {	// scan for max. volume
					cur_volume = ((short*)outbuf)[j]  ;
					if (cur_volume > max_volume ) max_volume = cur_volume;
				}
			}
			do	// decode rest of one input
			{
				ret = decodeMP3(&mp,NULL,0,outbuf,8192,&size);
				if (ret==MP3_OK) {	//size only defined for OK-frame. if not MP3_OK buffer is done
					byte_total = byte_total+size;
					for (int j = 0; j<size/2; j++) {	// scan for max. volume
						cur_volume = ((short*)outbuf)[j]  ;
						if (cur_volume > max_volume ) max_volume = cur_volume;
					}
				}
			}
			while (ret==MP3_OK);
		}
		// setting videoinfo-properties 
		vi.audio_samples_per_second = freqs[mp.fr.sampling_frequency];
		(mp.fr.stereo==2) ? channels=2 : channels = 1;

		if (mp.fr.stereo==2) {
			channels = 2;
			vi.num_audio_samples = byte_total / 4;
			vi.stereo = true;
		}
		else {
			channels = 1;
			vi.num_audio_samples = byte_total / 2;
			vi.stereo = false;
		}
		last_sample = vi.num_audio_samples;	//init done, now at end-of-file
		vi.num_frames = vi.FramesFromAudioSamples(vi.num_audio_samples);

	}
	else {	// seek, decode and write PCM to AviSynth buffer
		if (sample_number > vi.num_audio_samples)
			return 1; // No scan over max position possible

		if (sample_number < last_sample) 	// go to start if trying to seek backward
		{
			InitMP3(&mp);	//clear mp structure
			mpfile.clear();	//reset EOF flag
			mpfile.seekg(0);
			do // read and decode first input buffer. can be necessary more than once to fill output buffer
			{
				mpfile.read(inbuf,16384);
				len = mpfile.gcount();
				ret = decodeMP3(&mp,inbuf,len,outbuf,8192,&size);
			} while (ret==MP3_NEED_MORE);
			last_sample = 0;
			last_byte = 0;
		}
		
		(last_sample - sample_number > 32768) ? mp.quick_dirty = true : mp.quick_dirty = false;	//fast searching on/off
		
		while (sample_number > last_sample) 
		{
			last_sample++;
			last_byte = last_byte + channels * 2;

			ret = MP3_OK;
			if (last_byte > size-1)
			{
				ret = decodeMP3(&mp,NULL,0,outbuf,8192,&size);
				last_byte = 0;
			}
			if (ret!=MP3_OK) 
			{
				do
				{
					mpfile.read(inbuf,16384);
					len = mpfile.gcount();
					ret = decodeMP3(&mp,inbuf,len,outbuf,8192,&size);
				} while (ret==MP3_NEED_MORE);
				last_byte = 0;
			}
			if (ret!=MP3_OK) return 2;
		}
		for (int i = 0; i<channels; i++)
		{
			if (normalize) {
			((short*)pcm)[sample_offset*channels+i] = Saturate(32767.0/(double)max_volume * (double)((short*)outbuf)[last_byte/2+i]);
			} else {
			((short*)pcm)[sample_offset*channels+i] = ((short*)outbuf)[last_byte/2+i];
			};

		}
	};
	return 0;
}

MPASource::MPASource(const char filename[], const bool _normalize, IScriptEnvironment* env) {
		vi.width = 8;	//generate a small video. helpful for testing
        vi.height = 8;
        vi.fps_numerator = 25;
        vi.fps_denominator = 1;
        vi.pixel_type = VideoInfo::YUY2;
		vi.sixteen_bit = true;

	normalize = _normalize;

	mpfile.open (filename, ios::in | ios::binary | ios::nocreate );
	if (!mpfile.is_open()) {
		env->ThrowError("File could not be opened!");
	}
	
	GetSample(NULL, 0, 0);	//init vi

}

MPASource::~MPASource(){
	ExitMP3(&mp);
}


void __stdcall MPASource::GetAudio(void* buf, int start, int count, IScriptEnvironment* env) {

	for (int i=0;i<(count);i++){
		GetSample(buf, i, start+i);
	}
}


PVideoFrame __stdcall MPASource::GetFrame(int n, IScriptEnvironment* env) {

    PVideoFrame dst = env->NewVideoFrame(vi);
    BYTE * yuy2p = dst->GetWritePtr();
    const int dst_pitch = dst->GetPitch();

    for (int y = 0; y < vi.height; y++) {
        for (int x = 0; x < vi.width/2; x++) {	// each x is a double pixel in YUY2
            yuy2p[x*4 + 0] = 128;	//Y
            yuy2p[x*4 + 1] = 128; //U
            yuy2p[x*4 + 2] = 128; //Y
            yuy2p[x*4 + 3] = 128; //Y
        }
        yuy2p += dst_pitch;
    }
    return dst;
}


bool __stdcall MPASource::GetParity(int n) { return false; }
const VideoInfo & __stdcall MPASource::GetVideoInfo() { return vi; }

AVSValue __cdecl Create_MPASource(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new MPASource(args[0].AsString(""), args[1].AsBool(false), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env)
{
  env->AddFunction("MPASource","s[normalize]b",Create_MPASource,0);
  return 0;
}
